import {
  Packwiz,
  Resource,
  type Metafile,
  type PackwizIndex,
  HashFormat,
  Side,
  isValidSide,
  doHashesMatch,
} from "@packwizjs/parser";
import { write } from "bun";

interface HashInfo {
  hashFormat: HashFormat;
  hash: string;
}

interface State {
  stateVersion: number;
  currentVersion: string;
  hashes: {
    [fileName: string]: HashInfo;
  };
}

function getSaveLocation(
  path: Resource,
  index: PackwizIndex,
  metafile?: Metafile,
): Resource {
  const cwd = new Resource(process.cwd());
  const diff = index.location.diff(path);
  const fileDirectory = cwd.join(...diff);
  if (metafile) {
    return fileDirectory.parent.join(metafile.filename);
  } else {
    return fileDirectory;
  }
}

async function writeStateFile(
  packwiz: Packwiz,
  hashes: { [fileName: string]: HashInfo },
) {
  const stateFile = packwiz.index.location.parent.join(".packwizjs-state.json");
  const state: State = {
    stateVersion: 1,
    currentVersion: packwiz.version,
    hashes: hashes,
  };
  // for (const file of packwiz.index.files) {
  //   state.hashes[packwiz.index.location.diff(file.file).join("/")] = {
  //     hashFormat: file.hashFormat,
  //     hash: file.hash,
  //   };
  // }
  const saveLocation = getSaveLocation(stateFile, packwiz.index);
  await write(saveLocation.toString(), JSON.stringify(state, null));
}

/**
 * Reads the state file from the specified location.
 * @param stateFile - The path to the state file. Defaults to ".packwizjs-state.json" in the current working directory.
 * @returns A Promise containing the parsed state object.
 */
export async function readStateFile(
  stateFile: Resource = new Resource(process.cwd()).join(
    ".packwizjs-state.json",
  ),
): Promise<State> {
  // When adding a new state version, do NOT change the name of the stateVersion key!!! That key is NOT versioned.
  const json = JSON.parse(await stateFile.fetchContents());
  const stateVersion: number = json["stateVersion"];
  switch (stateVersion) {
    case 1:
      const state: State = { ...json };
      return state;
    default:
      throw new Error(
        `Unsupported state version: ${stateVersion}. Please update '@packwizjs/sync' to use newer state file versions.`,
      );
  }
}

async function downloadFile(
  hash: string,
  hashFormat: HashFormat,
  url: Resource,
  path: Resource,
) {
  const response = await fetch(new URL(url.toString()));
  if (!response.ok) {
    throw new Error(`Failed to download ${url}: ${response.statusText}`);
  }
  const arrayBuffer = await response.arrayBuffer();
  if (arrayBuffer.byteLength === 0) {
    console.warn(`WARNING: Downloaded file is empty: ${url}`);
  }
  const dataBuffer = Buffer.from(new Uint8Array(arrayBuffer));
  doHashesMatch(hash, hashFormat, dataBuffer);
  console.log(`Saving file to ${path.toString()}`);
  const fileSize = await write(path.toString(), arrayBuffer);
  console.log(`Saved ${formatBytes(fileSize)} file to ${path.toString()}`);
}

/**
 * Converts bytes into a human-readable format with size units.
 * @param bytes - The number of bytes.
 * @param decimals - The number of decimal places to use.
 * @returns Formatted string (e.g., "1.23 MB").
 */
export function formatBytes(bytes: number, decimals: number = 2) {
  if (bytes === 0) return "0 Bytes";
  const k = 1024; // or 1000 for decimal units
  const dm = decimals < 0 ? 0 : decimals;
  const sizes = ["Bytes", "KB", "MB", "GB", "TB", "PB", "EB", "ZB", "YB"];
  const i = Math.floor(Math.log(bytes) / Math.log(k));
  return parseFloat((bytes / Math.pow(k, i)).toFixed(dm)) + " " + sizes[i];
}

/**
 * Iterates over the files in a Packwiz index and downloads them concurrently.
 * @param index The Packwiz index to iterate over.
 * @param side The side to download files for (e.g., "client", "server", or "both").
 * @param path The path to save the downloaded files. Defaults to the current working directory.
 * @param concurrencyLimit The maximum number of concurrent downloads.
 * @returns A promise that resolves when all files have been downloaded.
 */
export async function iteratePackwizIndex(
  packwiz: Packwiz,
  side: Side,
  path: Resource = new Resource(process.cwd()),
  concurrencyLimit: number = 5,
) {
  let currentIndex = 0;
  let activeDownloads = 0;
  const totalFiles = packwiz.index.files.length;
  const stateFile = path.join(".packwizjs-state.json");
  let state: State | undefined;
  if (await stateFile.exists()) {
    state = await readStateFile(stateFile);
  }
  console.log(state);
  side = side.toLowerCase() as Side;
  const hashes: { [fileName: string]: HashInfo } = {};

  return new Promise<void>((resolve, reject) => {
    function downloadNextFile() {
      // Resolve if all file entries are downloaded and no active downloads are ongoing
      if (currentIndex >= totalFiles && activeDownloads === 0) {
        resolve();
        return;
      }

      // If there is still file entries to download and there are empty concurrency slots,
      // start a new download.
      while (activeDownloads < concurrencyLimit && currentIndex < totalFiles) {
        const file = packwiz.index.files[currentIndex++];
        activeDownloads++;

        (async () => {
          try {
            let saveLocation: Resource;
            let url: Resource;
            let hash: string;
            let hashFormat: HashFormat;

            if (file.metafile) {
              const metafile = await file.parse();

              if (!metafile.side || !isValidSide(metafile.side)) {
                throw new Error(
                  `Metafile ${metafile.filename} has invalid Side: ${metafile.side}`,
                );
              }

              if (metafile.side !== Side.Both && metafile.side !== side) {
                // Don't throw an error here, this is normal behavior!
                console.log(
                  `Skipping metafile ${metafile.filename} due to side mismatch: '${metafile.side}' (Wanted '${side}' or '${Side.Both}')`,
                );
                return;
              }

              hash = metafile.provider.hash;
              hashFormat = metafile.provider.hashFormat;
              url = metafile.provider.url;
              saveLocation = getSaveLocation(
                file.file,
                packwiz.index,
                metafile,
              );
            } else {
              // we don't check non-metafiles' Side because packwiz doesn't store metadata for whether or not to download them on client / server
              // so instead, always download them on both sides (assume Side.Both)
              const diff = packwiz.index.location.diff(file.file);
              hash = file.hash;
              hashFormat = file.hashFormat;
              url = packwiz.index.location.parent.join(...diff);
              saveLocation = getSaveLocation(file.file, packwiz.index);
            }

            if (state) {
              let stateHash: HashInfo | undefined;
              // TODO: implement hash checking for files downloaded via metafile providers
              if (!file.metafile) {
                stateHash =
                  state.hashes[
                    packwiz.index.location.diff(file.file).join("/")
                  ];
              }
              if (stateHash?.hash && stateHash.hash === hash) {
                console.log(
                  `Skipping already downloaded file ${file.file.toString()}`,
                );
              } else {
                await downloadFile(hash, hashFormat, url, saveLocation);
              }
            } else {
              await downloadFile(hash, hashFormat, url, saveLocation);
            }

            hashes[packwiz.index.location.diff(file.file).join("/")] = {
              hashFormat: hashFormat,
              hash: hash,
            };
          } catch (error) {
            console.error(`Error downloading file: ${error}`);
            reject(error);
            return;
          } finally {
            activeDownloads--;
            downloadNextFile();
          }
        })();
      }
    }
    downloadNextFile();
  }).then(() => writeStateFile(packwiz, hashes));
}
